@@ -40,7 +40,7 @@ device_set_property (XDevice        *xdevice,
                      const char     *device_name,
                      PropertyHelper *property)
 {
-        int rc;
+        int rc, i;
         Atom prop;
         Atom realtype;
         int realformat;
@@ -56,28 +56,33 @@ device_set_property (XDevice        *xdevice,
 
         rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
                                  xdevice, prop, 0, property->nitems, False,
-                                 XA_INTEGER, &realtype, &realformat, &nitems,
+                                 AnyPropertyType, &realtype, &realformat, &nitems,
                                  &bytes_after, &data);
 
-        if (rc == Success && realtype == XA_INTEGER &&
-            realformat == property->format && nitems >= property->nitems) {
-                int i;
-                for (i = 0; i < nitems; i++) {
-                        switch (property->format) {
-                                case 8:
-                                        data[i] = property->data.c[i];
-                                        break;
-                                case 32:
-                                        ((long*)data)[i] = property->data.i[i];
-                                        break;
-                        }
-                }
-
-                XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
-                                       xdevice, prop, XA_INTEGER, realformat,
-                                       PropModeReplace, data, nitems);
+        if (rc != Success ||
+            realtype != property->type ||
+            realformat != property->format ||
+            nitems < property->nitems) {
+                gdk_error_trap_pop_ignored ();
+                g_warning ("Error reading property \"%s\" for \"%s\"", property->name, device_name);
+                return FALSE;
         }
 
+        for (i = 0; i < nitems; i++) {
+                switch (property->format) {
+                        case 8:
+                                data[i] = property->data.c[i];
+                                break;
+                        case 32:
+                                ((long*)data)[i] = property->data.i[i];
+                                break;
+                }
+        }
+
+        XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
+                               xdevice, prop, realtype, realformat,
+                               PropModeReplace, data, nitems);
+
         if (gdk_error_trap_pop ()) {
                 g_warning ("Error in setting \"%s\" for \"%s\"", property->name, device_name);
                 return FALSE;
@@ -86,24 +91,35 @@ device_set_property (XDevice        *xdevice,
         return TRUE;
 }
 
-gboolean
-supports_xinput_devices (void)
+static gboolean
+supports_xinput_devices_with_opcode (int *opcode)
 {
         gint op_code, event, error;
+        gboolean retval;
 
-        return XQueryExtension (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
-                                "XInputExtension",
-                                &op_code,
-                                &event,
-                                &error);
+        retval = XQueryExtension (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
+				  "XInputExtension",
+				  &op_code,
+				  &event,
+				  &error);
+	if (opcode)
+		*opcode = op_code;
+
+	return retval;
 }
 
 gboolean
-supports_xinput2_devices (void)
+supports_xinput_devices (void)
+{
+	return supports_xinput_devices_with_opcode (NULL);
+}
+
+gboolean
+supports_xinput2_devices (int *opcode)
 {
         int major, minor;
 
-        if (supports_xinput_devices () == FALSE)
+        if (supports_xinput_devices_with_opcode (opcode) == FALSE)
                 return FALSE;
 
         gdk_error_trap_push ();
@@ -225,6 +241,131 @@ touchpad_is_present (void)
                                        device_is_touchpad);
 }
 
+char *
+xdevice_get_device_node (int deviceid)
+{
+        Atom           prop;
+        Atom           act_type;
+        int            act_format;
+        unsigned long  nitems, bytes_after;
+        unsigned char *data;
+        char          *ret;
+
+        gdk_display_sync (gdk_display_get_default ());
+
+        prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Device Node", False);
+        if (!prop)
+                return NULL;
+
+        gdk_error_trap_push ();
+
+        if (!XIGetProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
+                            deviceid, prop, 0, 1000, False,
+                            AnyPropertyType, &act_type, &act_format,
+                            &nitems, &bytes_after, &data) == Success) {
+                gdk_error_trap_pop_ignored ();
+                return NULL;
+        }
+        if (gdk_error_trap_pop ())
+                goto out;
+
+        if (nitems == 0)
+                goto out;
+
+        if (act_type != XA_STRING)
+                goto out;
+
+        /* Unknown string format */
+        if (act_format != 8)
+                goto out;
+
+        ret = g_strdup ((char *) data);
+
+        XFree (data);
+        return ret;
+
+out:
+        XFree (data);
+        return NULL;
+}
+
+#define TOOL_ID_FORMAT_SIZE 32
+static int
+get_id_for_index (guchar *data,
+		  guint   idx)
+{
+	guchar *ptr;
+	int id;
+
+	ptr = data;
+	ptr += TOOL_ID_FORMAT_SIZE / 8 * idx;
+
+	id = *((int32_t*)ptr);
+	id = id & 0xfffff;
+
+	return id;
+}
+
+
+#define STYLUS_DEVICE_ID        0x02
+#define ERASER_DEVICE_ID        0x0A
+
+int
+xdevice_get_last_tool_id (int  deviceid)
+{
+        Atom           prop;
+        Atom           act_type;
+        int            act_format;
+        unsigned long  nitems, bytes_after;
+        unsigned char *data;
+        int            id;
+
+        id = -1;
+
+        gdk_display_sync (gdk_display_get_default ());
+
+        prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), WACOM_SERIAL_IDS_PROP, False);
+        if (!prop)
+                return -1;
+
+        gdk_error_trap_push ();
+
+        if (!XIGetProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
+                            deviceid, prop, 0, 1000, False,
+                            AnyPropertyType, &act_type, &act_format,
+                            &nitems, &bytes_after, &data) == Success) {
+                gdk_error_trap_pop_ignored ();
+                return -1;
+        }
+
+        if (gdk_error_trap_pop ())
+                goto out;
+
+	if (nitems != 4)
+		goto out;
+
+	if (act_type != XA_INTEGER)
+		goto out;
+
+	if (act_format != TOOL_ID_FORMAT_SIZE)
+		goto out;
+
+	/* item 0 = tablet ID
+	 * item 1 = old device serial number (== last tool in proximity)
+	 * item 2 = old hardware serial number (including tool ID)
+	 * item 3 = current serial number (0 if no tool in proximity) */
+	id = get_id_for_index (data, 2);
+
+	/* That means that no tool was set down yet */
+	if (id == STYLUS_DEVICE_ID ||
+	    id == ERASER_DEVICE_ID)
+		return 0x0;
+
+out:
+        XFree (data);
+        return id;
+}
+
 gboolean
 set_device_enabled (int device_id,
                     gboolean enabled)
