@@ -32,8 +32,10 @@
 
 #include <glib.h>
 #include <glib/gi18n.h>
+#include <glib/gstdio.h>
 
 #include <cups/cups.h>
+#include <cups/ppd.h>
 #include <libnotify/notify.h>
 
 #include "gnome-settings-profile.h"
@@ -260,9 +262,9 @@ on_cups_notification (GDBusConnection *connection,
         GsdPrintNotificationsManager *manager = (GsdPrintNotificationsManager *) user_data;
         gboolean                     printer_is_accepting_jobs;
         gboolean                     my_job = FALSE;
+        gboolean                     known_reason;
         http_t                      *http;
         gchar                       *printer_name = NULL;
-        gchar                       *display_name = NULL;
         gchar                       *primary_text = NULL;
         gchar                       *secondary_text = NULL;
         gchar                       *text = NULL;
@@ -436,9 +438,6 @@ on_cups_notification (GDBusConnection *connection,
                 g_hash_table_remove (manager->priv->printing_printers,
                                      printer_name);
 
-                /* FIXME: get a better human readable name */
-                display_name = g_strdup (printer_name);
-
                 switch (job_state) {
                         case IPP_JOB_PENDING:
                         case IPP_JOB_HELD:
@@ -448,33 +447,28 @@ on_cups_notification (GDBusConnection *connection,
                                 /* Translators: A print job has been stopped */
                                 primary_text = g_strdup (_("Printing stopped"));
                                 /* Translators: "print-job xy" on a printer */
-                                secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, display_name);
+                                secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, printer_name);
                                 break;
                         case IPP_JOB_CANCELED:
                                 /* Translators: A print job has been canceled */
                                 primary_text = g_strdup (_("Printing canceled"));
                                 /* Translators: "print-job xy" on a printer */
-                                secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, display_name);
+                                secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, printer_name);
                                 break;
                         case IPP_JOB_ABORTED:
                                 /* Translators: A print job has been aborted */
                                 primary_text = g_strdup (_("Printing aborted"));
                                 /* Translators: "print-job xy" on a printer */
-                                secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, display_name);
+                                secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, printer_name);
                                 break;
                         case IPP_JOB_COMPLETED:
                                 /* Translators: A print job has been completed */
                                 primary_text = g_strdup (_("Printing completed"));
                                 /* Translators: "print-job xy" on a printer */
-                                secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, display_name);
+                                secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, printer_name);
                                 break;
                 }
         } else if (g_strcmp0 (signal_name, "JobState") == 0 && my_job) {
-                display_name = get_dest_attr (printer_name,
-                                              "printer-info",
-                                              manager->priv->dests,
-                                              manager->priv->num_dests);
-
                 switch (job_state) {
                         case IPP_JOB_PROCESSING:
                                 g_hash_table_insert (manager->priv->printing_printers,
@@ -483,7 +477,7 @@ on_cups_notification (GDBusConnection *connection,
                                 /* Translators: A job is printing */
                                 primary_text = g_strdup (_("Printing"));
                                 /* Translators: "print-job xy" on a printer */
-                                secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, display_name);
+                                secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, printer_name);
                                 break;
                         case IPP_JOB_STOPPED:
                         case IPP_JOB_CANCELED:
@@ -496,11 +490,6 @@ on_cups_notification (GDBusConnection *connection,
                                 break;
                 }
         } else if (g_strcmp0 (signal_name, "JobCreated") == 0 && my_job) {
-                display_name = get_dest_attr (printer_name,
-                                              "printer-info",
-                                              manager->priv->dests,
-                                              manager->priv->num_dests);
-
                 if (job_state == IPP_JOB_PROCESSING) {
                         g_hash_table_insert (manager->priv->printing_printers,
                                              g_strdup (printer_name), NULL);
@@ -508,7 +497,7 @@ on_cups_notification (GDBusConnection *connection,
                         /* Translators: A job is printing */
                         primary_text = g_strdup (_("Printing"));
                         /* Translators: "print-job xy" on a printer */
-                        secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, display_name);
+                        secondary_text = g_strdup_printf (_("\"%s\" on %s"), job_name, printer_name);
                 }
         } else if (g_strcmp0 (signal_name, "PrinterStateChanged") == 0) {
                 cups_dest_t  *dest = NULL;
@@ -567,9 +556,6 @@ on_cups_notification (GDBusConnection *connection,
 
                 /* Check whether we are printing on this printer right now. */
                 if (g_hash_table_lookup_extended (manager->priv->printing_printers, printer_name, NULL, NULL)) {
-                        /* FIXME: get a better human readable name */
-                        display_name = g_strdup (printer_name);
-
                         dest = cupsGetDest (printer_name,
                                             NULL,
                                             manager->priv->num_dests,
@@ -630,11 +616,13 @@ on_cups_notification (GDBusConnection *connection,
 
                         for (tmp_list = added_reasons; tmp_list; tmp_list = tmp_list->next) {
                                 gchar *data = (gchar *) tmp_list->data;
+                                known_reason = FALSE;
                                 for (j = 0; j < G_N_ELEMENTS (reasons); j++) {
                                         if (strncmp (data,
                                                      reasons[j],
                                                      strlen (reasons[j])) == 0) {
                                                 NotifyNotification *notification;
+                                                known_reason = TRUE;
 
                                                 if (g_strcmp0 (reasons[j], "connecting-to-device") == 0) {
                                                         TimeoutData *data;
@@ -682,6 +670,94 @@ on_cups_notification (GDBusConnection *connection,
                                                 }
                                         }
                                 }
+
+                                if (!known_reason && !g_str_equal (data, "none")) {
+                                        NotifyNotification *notification;
+                                        ReasonData         *reason_data;
+                                        gchar              *first_row;
+                                        gchar              *second_row;
+                                        gchar              *text = NULL;
+                                        gchar              *ppd_file_name;
+                                        ppd_file_t         *ppd_file;
+                                        char                buffer[8192];
+
+                                        ppd_file_name = g_strdup (cupsGetPPD (printer_name));
+                                        if (ppd_file_name) {
+                                                ppd_file = ppdOpenFile (ppd_file_name);
+                                                if (ppd_file) {
+                                                        gchar **tmpv;
+                                                        static const char * const schemes[] = {
+                                                                "text", "http", "help", "file"
+                                                        };
+
+                                                        tmpv = g_new0 (gchar *, G_N_ELEMENTS (schemes) + 1);
+                                                        i = 0;
+                                                        for (j = 0; j < G_N_ELEMENTS (schemes); j++) {
+                                                                if (ppdLocalizeIPPReason (ppd_file, data, schemes[j], buffer, sizeof (buffer))) {
+                                                                        tmpv[i++] = g_strdup (buffer);
+                                                                }
+                                                        }
+
+                                                        if (i > 0)
+                                                                text = g_strjoinv (", ", tmpv);
+                                                        g_strfreev (tmpv);
+
+                                                        ppdClose (ppd_file);
+                                                }
+
+                                                g_unlink (ppd_file_name);
+                                                g_free (ppd_file_name);
+                                        }
+
+
+                                        if (g_str_has_suffix (data, "-report"))
+                                                /* Translators: This is a title of a report notification for a printer */
+                                                first_row = g_strdup (_("Printer report"));
+                                        else if (g_str_has_suffix (data, "-warning"))
+                                                /* Translators: This is a title of a warning notification for a printer */
+                                                first_row = g_strdup (_("Printer warning"));
+                                        else
+                                                /* Translators: This is a title of an error notification for a printer */
+                                                first_row = g_strdup (_("Printer error"));
+
+
+                                        if (text == NULL)
+                                                text = g_strdup (data);
+
+                                        /* Translators: "Printer 'MyPrinterName': 'Description of the report/warning/error from a PPD file'." */
+                                        second_row = g_strdup_printf (_("Printer '%s': '%s'."), printer_name, text);
+                                        g_free (text);
+
+
+                                        notification = notify_notification_new (first_row,
+                                                                                second_row,
+                                                                                "printer-symbolic");
+                                        notify_notification_set_app_name (notification, _("Printers"));
+                                        notify_notification_set_hint (notification,
+                                                                      "resident",
+                                                                      g_variant_new_boolean (TRUE));
+                                        notify_notification_set_timeout (notification, REASON_TIMEOUT);
+
+                                        reason_data = g_new0 (ReasonData, 1);
+                                        reason_data->printer_name = g_strdup (printer_name);
+                                        reason_data->reason = g_strdup (data);
+                                        reason_data->notification = notification;
+                                        reason_data->manager = manager;
+
+                                        reason_data->notification_close_id =
+                                                g_signal_connect (notification,
+                                                                  "closed",
+                                                                  G_CALLBACK (notification_closed_cb),
+                                                                  reason_data);
+
+                                        manager->priv->active_notifications =
+                                                g_list_append (manager->priv->active_notifications, reason_data);
+
+                                        notify_notification_show (notification, NULL);
+
+                                        g_free (first_row);
+                                        g_free (second_row);
+                                }
                         }
                         g_slist_free (added_reasons);
                 }
@@ -693,8 +769,6 @@ on_cups_notification (GDBusConnection *connection,
                         g_strfreev (old_state_reasons);
         }
 
-        g_free (display_name);
-
 
         if (primary_text) {
                 NotifyNotification *notification;
