@@ -35,9 +35,13 @@
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
 #include <X11/Xatom.h>
+#include <X11/extensions/XTest.h>
+#include <X11/keysym.h>
+#include <Xwacom.h>
 
 #include "gsd-enums.h"
 #include "gsd-input-helper.h"
+#include "gsd-keygrab.h"
 #include "gnome-settings-profile.h"
 #include "gsd-wacom-manager.h"
 #include "gsd-wacom-device.h"
@@ -49,7 +53,6 @@
 #define KEY_TPCBUTTON           "tablet-pc-button"
 #define KEY_IS_ABSOLUTE         "is-absolute"
 #define KEY_AREA                "area"
-#define KEY_PAD_BUTTON_MAPPING  "pad-buttonmapping"
 #define KEY_DISPLAY             "display"
 
 /* Stylus and Eraser settings */
@@ -57,6 +60,10 @@
 #define KEY_PRESSURETHRESHOLD   "pressurethreshold"
 #define KEY_PRESSURECURVE       "pressurecurve"
 
+/* Button settings */
+#define KEY_ACTION_TYPE         "action-type"
+#define KEY_CUSTOM_ACTION       "custom-action"
+
 /* See "Wacom Pressure Threshold" */
 #define DEFAULT_PRESSURE_THRESHOLD 27
 
@@ -67,6 +74,10 @@ struct GsdWacomManagerPrivate
         guint device_added_id;
         guint device_removed_id;
         GHashTable *devices;
+
+        /* button capture */
+        GSList *screens;
+        int      opcode;
 };
 
 static void     gsd_wacom_manager_class_init  (GsdWacomManagerClass *klass);
@@ -109,17 +120,28 @@ gsd_wacom_manager_class_init (GsdWacomManagerClass *klass)
         g_type_class_add_private (klass, sizeof (GsdWacomManagerPrivate));
 }
 
-static XDevice *
-open_device (GsdWacomDevice *device)
+static int
+get_device_id (GsdWacomDevice *device)
 {
-	XDevice *xdev;
 	GdkDevice *gdk_device;
 	int id;
 
 	g_object_get (device, "gdk-device", &gdk_device, NULL);
 	if (gdk_device == NULL)
-		return NULL;
+		return -1;
 	g_object_get (gdk_device, "device-id", &id, NULL);
+	return id;
+}
+
+static XDevice *
+open_device (GsdWacomDevice *device)
+{
+	XDevice *xdev;
+	int id;
+
+	id = get_device_id (device);
+	if (id < 0)
+		return NULL;
 
 	gdk_error_trap_push ();
 	xdev = XOpenDevice (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), id);
@@ -271,7 +293,8 @@ set_device_buttonmap (GsdWacomDevice *device,
 		g_usleep (300);
 	}
 
-	if (gdk_error_trap_pop () || rc != Success)
+	if ((gdk_error_trap_pop () && rc != MappingSuccess) ||
+	    rc != MappingSuccess)
 		g_warning ("Error in setting button mapping for \"%s\"", gsd_wacom_device_get_tool_name (device));
 
 	g_free (map);
@@ -352,6 +375,103 @@ apply_stylus_settings (GsdWacomDevice *device)
 	set_pressurethreshold (device, threshold);
 }
 
+static struct {
+	const char *button;
+	int         num;
+} def_buttons[] = {
+	/* Touchrings */
+	{ "AbsWheelUp", 90 },
+	{ "AbsWheelDown", 91 },
+	{ "AbsWheel2Up", 92 },
+	{ "AbsWheel2Down", 93 },
+	/* Touchstrips */
+	{ "StripLeftUp", 94 },
+	{ "StripLeftDown", 95 },
+	{ "StripRightUp", 96 },
+	{ "StripRightDown", 97 }
+};
+#define NUM_TOUCH_BUTTONS 4
+
+static void
+reset_touch_buttons (XDevice    *xdev,
+		     guint       offset,
+		     const char *device_property)
+{
+	Atom actions[NUM_TOUCH_BUTTONS];
+	Atom action_prop;
+	guint i;
+
+	/* Create a device property with the action for button i */
+	for (i = 0; i < NUM_TOUCH_BUTTONS; i++)
+	{
+		char *propname;
+		int action[2]; /* press + release */
+		Atom prop;
+		int mapped_button = def_buttons[i + offset].num;
+
+		action[0] = AC_BUTTON | mapped_button;
+
+		propname = g_strdup_printf ("Button %s action", def_buttons[i + offset].button);
+		prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), propname, False);
+		g_free (propname);
+		XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev,
+				       prop, XA_INTEGER, 32, PropModeReplace,
+				       (const guchar *) &action, 2);
+
+		/* prop now contains press + release for the mapped button */
+		actions[i] = prop;
+	}
+
+	/* Now set the actual action property to contain references to the various
+	 * actions */
+	action_prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), device_property, True);
+	XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev,
+			       action_prop, XA_ATOM, 32, PropModeReplace,
+			       (const guchar *) actions, NUM_TOUCH_BUTTONS);
+}
+
+static void
+reset_pad_buttons (GsdWacomDevice *device)
+{
+	XDevice *xdev;
+	int nmap;
+	unsigned char *map;
+	int i, j, rc;
+
+	/* Normal buttons */
+	xdev = open_device (device);
+
+	gdk_error_trap_push ();
+
+	nmap = 256;
+	map = g_new0 (unsigned char, nmap);
+	for (i = 0; i < nmap && i < sizeof (map); i++)
+		map[i] = i + 1;
+
+	/* X refuses to change the mapping while buttons are engaged,
+	 * so if this is the case we'll retry a few times */
+	for (j = 0;
+	     j < 20 && (rc = XSetDeviceButtonMapping (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev, map, nmap)) == MappingBusy;
+	     ++j) {
+		g_usleep (300);
+	}
+
+	if ((gdk_error_trap_pop () && rc != MappingSuccess) ||
+	    rc != MappingSuccess)
+		g_warning ("Error in resetting button mapping for \"%s\" (rc=%d)", gsd_wacom_device_get_tool_name (device), rc);
+
+	g_free (map);
+
+	gdk_error_trap_push ();
+	reset_touch_buttons (xdev, 0, "Wacom Wheel Buttons");
+	reset_touch_buttons (xdev, 4, "Wacom Strip Buttons");
+	gdk_error_trap_pop_ignored ();
+
+	XCloseDevice (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev);
+
+	/* FIXME, set the LED(s) for the mode(s) too */
+}
+
 static void
 set_wacom_settings (GsdWacomManager *manager,
 		    GsdWacomDevice  *device)
@@ -384,7 +504,11 @@ set_wacom_settings (GsdWacomManager *manager,
 	}
 
 	if (type == WACOM_TYPE_PAD) {
-		set_device_buttonmap (device, g_settings_get_value (settings, KEY_PAD_BUTTON_MAPPING));
+		int id;
+
+		id = get_device_id (device);
+		reset_pad_buttons (device);
+		grab_button (id, TRUE, manager->priv->screens);
 		return;
 	}
 
@@ -425,9 +549,6 @@ wacom_settings_changed (GSettings      *settings,
 		if (type != WACOM_TYPE_CURSOR &&
 		    type != WACOM_TYPE_PAD)
 			set_area (device, g_settings_get_value (settings, key));
-	} else if (g_str_equal (key, KEY_PAD_BUTTON_MAPPING)) {
-		if (type == WACOM_TYPE_PAD)
-			set_device_buttonmap (device, g_settings_get_value (settings, key));
 	} else if (g_str_equal (key, KEY_DISPLAY)) {
 		if (type != WACOM_TYPE_CURSOR &&
 		    type != WACOM_TYPE_PAD)
@@ -534,6 +655,153 @@ device_removed_cb (GdkDeviceManager *device_manager,
 	g_hash_table_remove (manager->priv->devices, gdk_device);
 }
 
+static GsdWacomDevice *
+device_id_to_device (GsdWacomManager *manager,
+		     int              deviceid)
+{
+	GList *devices, *l;
+	GsdWacomDevice *ret;
+
+	ret = NULL;
+	devices = g_hash_table_get_keys (manager->priv->devices);
+
+	for (l = devices; l != NULL; l = l->next) {
+		GdkDevice *device = l->data;
+		int id;
+
+		g_object_get (device, "device-id", &id, NULL);
+		if (id == deviceid) {
+			ret = g_hash_table_lookup (manager->priv->devices, device);
+			break;
+		}
+	}
+
+	g_list_free (devices);
+	return ret;
+}
+
+struct {
+	guint mask;
+	KeySym keysym;
+} mods_keysyms[] = {
+	{ GDK_MOD1_MASK, XK_Alt_L },
+	{ GDK_SHIFT_MASK, XK_Shift_L },
+	{ GDK_CONTROL_MASK, XK_Control_L },
+};
+
+static void
+send_modifiers (Display *display,
+		XDevice *dev,
+		guint mask,
+		gboolean is_press)
+{
+	guint i;
+
+	for (i = 0; i < G_N_ELEMENTS(mods_keysyms); i++) {
+		if (mask & mods_keysyms[i].mask) {
+			guint keycode;
+
+			keycode = XKeysymToKeycode (display, mods_keysyms[i].keysym);
+			XTestFakeDeviceKeyEvent (display, dev, keycode,
+						 is_press ? True : False, NULL, 0, 0);
+		}
+	}
+}
+
+static GdkFilterReturn
+filter_button_events (XEvent          *xevent,
+                      GdkEvent        *event,
+                      GsdWacomManager *manager)
+{
+	XIEvent             *xiev;
+	XIDeviceEvent       *xev;
+	XGenericEventCookie *cookie;
+	guint                deviceid;
+	GsdWacomDevice      *device;
+	int                  button;
+	GsdWacomTabletButton *wbutton;
+	GtkDirectionType      dir;
+	XDevice               dev;
+	char                 *str;
+	guint                 keyval;
+	guint                *keycodes;
+	guint                 mods;
+
+        /* verify we have a key event */
+	if (xevent->type != GenericEvent)
+		return GDK_FILTER_CONTINUE;
+	cookie = &xevent->xcookie;
+	if (cookie->extension != manager->priv->opcode)
+		return GDK_FILTER_CONTINUE;
+
+	xiev = (XIEvent *) xevent->xcookie.data;
+
+	if (xiev->evtype != XI_ButtonRelease &&
+	    xiev->evtype != XI_ButtonPress)
+		return GDK_FILTER_CONTINUE;
+
+	xev = (XIDeviceEvent *) xiev;
+
+	deviceid = xev->sourceid;
+	device = device_id_to_device (manager, deviceid);
+	if (gsd_wacom_device_get_device_type (device) != WACOM_TYPE_PAD)
+		return GDK_FILTER_CONTINUE;
+
+	button = xev->detail;
+
+	/* FIXME, we'll also need to pass the current mode(s) */
+	wbutton = gsd_wacom_device_get_button (device, button, &dir);
+	if (wbutton == NULL) {
+		g_warning ("Could not find matching button for '%d' on '%s'",
+			   button, gsd_wacom_device_get_name (device));
+		return GDK_FILTER_CONTINUE;
+	}
+
+	g_debug ("Received event button '%s'%s ('%d') on device '%s' ('%d')",
+		 wbutton->id,
+		 wbutton->type == WACOM_TABLET_BUTTON_TYPE_ELEVATOR ?
+		 (dir == GTK_DIR_UP ? " 'up'" : " 'down'") : "",
+		 button,
+		 gsd_wacom_device_get_name (device),
+		 deviceid);
+
+	/* FIXME, don't know how to handle those yet */
+	if (wbutton->type == WACOM_TABLET_BUTTON_TYPE_ELEVATOR)
+		return GDK_FILTER_REMOVE;
+
+	/* FIXME, we need to switch mode here */
+	if (wbutton->type == WACOM_TABLET_BUTTON_TYPE_HARDCODED)
+		return GDK_FILTER_REMOVE;
+
+	/* Nothing to do */
+	if (g_settings_get_enum (wbutton->settings, KEY_ACTION_TYPE) == GSD_WACOM_ACTION_TYPE_NONE)
+		return GDK_FILTER_REMOVE;
+
+	str = g_settings_get_string (wbutton->settings, KEY_CUSTOM_ACTION);
+	gtk_accelerator_parse_with_keycode (str, &keyval, &keycodes, &mods);
+
+	if (keycodes == NULL) {
+		g_warning ("Failed to find a keycode for shortcut '%s'", str);
+		g_free (str);
+		return GDK_FILTER_REMOVE;
+	}
+	g_free (str);
+
+	dev.device_id = deviceid; /* that's cheating, but whot did it first */
+
+	/* And send out the keys! */
+	send_modifiers (xev->display, &dev, mods, TRUE);
+	XTestFakeDeviceKeyEvent (xev->display, &dev, keycodes[0],
+				 True, NULL, 0, 0);
+	XTestFakeDeviceKeyEvent (xev->display, &dev, keycodes[0],
+				 False, NULL, 0, 0);
+	send_modifiers (xev->display, &dev, mods, FALSE);
+
+	g_free (keycodes);
+
+	return GDK_FILTER_REMOVE;
+}
+
 static void
 set_devicepresence_handler (GsdWacomManager *manager)
 {
@@ -560,6 +828,7 @@ static gboolean
 gsd_wacom_manager_idle_cb (GsdWacomManager *manager)
 {
 	GList *devices, *l;
+	GSList *ls;
 
         gnome_settings_profile_start (NULL);
 
@@ -572,6 +841,13 @@ gsd_wacom_manager_idle_cb (GsdWacomManager *manager)
 		device_added_cb (manager->priv->device_manager, l->data, manager);
         g_list_free (devices);
 
+        /* Start filtering the button events */
+        for (ls = manager->priv->screens; ls != NULL; ls = ls->next) {
+                gdk_window_add_filter (gdk_screen_get_root_window (ls->data),
+                                       (GdkFilterFunc) filter_button_events,
+                                       manager);
+        }
+
         gnome_settings_profile_end (NULL);
 
         manager->priv->start_idle_id = 0;
@@ -579,17 +855,44 @@ gsd_wacom_manager_idle_cb (GsdWacomManager *manager)
         return FALSE;
 }
 
+static void
+init_screens (GsdWacomManager *manager)
+{
+        GdkDisplay *display;
+        int i;
+
+        display = gdk_display_get_default ();
+        for (i = 0; i < gdk_display_get_n_screens (display); i++) {
+                GdkScreen *screen;
+
+                screen = gdk_display_get_screen (display, i);
+                if (screen == NULL) {
+                        continue;
+                }
+                manager->priv->screens = g_slist_append (manager->priv->screens, screen);
+        }
+}
+
 gboolean
 gsd_wacom_manager_start (GsdWacomManager *manager,
                          GError         **error)
 {
+	int a, b, c, d;
+
         gnome_settings_profile_start (NULL);
 
-        if (supports_xinput2_devices (NULL) == FALSE) {
+        if (supports_xinput2_devices (&manager->priv->opcode) == FALSE) {
                 g_debug ("No Xinput2 support, disabling plugin");
                 return TRUE;
         }
 
+        if (!XTestQueryExtension (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), &a, &b, &c, &d)) {
+                g_debug ("No XTest extension support, disabling plugin");
+                return TRUE;
+        }
+
+        init_screens (manager);
+
         manager->priv->start_idle_id = g_idle_add ((GSourceFunc) gsd_wacom_manager_idle_cb, manager);
 
         gnome_settings_profile_end (NULL);
@@ -601,6 +904,7 @@ void
 gsd_wacom_manager_stop (GsdWacomManager *manager)
 {
         GsdWacomManagerPrivate *p = manager->priv;
+        GSList *ls;
 
         g_debug ("Stopping wacom manager");
 
@@ -609,6 +913,12 @@ gsd_wacom_manager_stop (GsdWacomManager *manager)
                 g_signal_handler_disconnect (p->device_manager, p->device_removed_id);
                 p->device_manager = NULL;
         }
+
+        for (ls = p->screens; ls != NULL; ls = ls->next) {
+                gdk_window_remove_filter (gdk_screen_get_root_window (ls->data),
+                                          (GdkFilterFunc) filter_button_events,
+                                          manager);
+        }
 }
 
 static void
@@ -628,6 +938,11 @@ gsd_wacom_manager_finalize (GObject *object)
                 wacom_manager->priv->devices = NULL;
         }
 
+        if (wacom_manager->priv->screens != NULL) {
+                g_slist_free (wacom_manager->priv->screens);
+                wacom_manager->priv->screens = NULL;
+        }
+
         if (wacom_manager->priv->start_idle_id != 0)
                 g_source_remove (wacom_manager->priv->start_idle_id);
 
